from pathlib import Path
from typing import List

from azure.ai.ml import command
from azure.ai.ml.entities import Environment

from azure_helper.interfaces.aml_interface import AMLInterface
from azure_helper.logger import get_logger

# from azureml.core import Experiment, ScriptRunConfig
# from azureml.core.runconfig import DockerConfiguration, RunConfiguration
# from azureml.pipeline.core import Pipeline
# from azureml.pipeline.steps import PythonScriptStep


log = get_logger()


class AMLExperiment:
    def __init__(
        self,
        aml_interface: AMLInterface,
        aml_compute_name: str,
        aml_compute_instance: str,
        env_name: str,
        experiment_name: str,
        training_script_path: str,
        min_node: int = 1,
        max_node: int = 2,
        clean_after_run: bool = True,
    ) -> None:
        """Instantiate the creation of an AzureML Experiment needed to train a model.

        Args:
            aml_interface (AMLInterface): The aml_interface needed to register the experiment in the workspace.
            aml_compute_name (str): The name of the compute instance used.
            aml_compute_instance (str): The size (as in `vm_size`) of the compute instance used.
            min_node (int, optional): The minimum number of nodes to use on the compute instance. Defaults to 1.
            max_node (int, optional): The maximum number of nodes to use on the compute instance. Defaults to 2.
            env_name (str): The name of the training environment.
            experiment_name (str): The name of the experiment.
            training_script_path (str): The path to the training loop script.
            clean_after_run (bool, optional): Whether or not you want to delete the compute instance after training. Defaults to True.
        """

        self.aml_interface = aml_interface
        self.aml_compute_name = aml_compute_name
        self.aml_compute_instance = aml_compute_instance
        self.env_name = env_name
        self.experiment_name = experiment_name
        self.clean_after_run = clean_after_run
        self.training_script_path = training_script_path

        # self.compute_target = aml_interface.get_compute_target(
        #     aml_compute_name,
        #     aml_compute_instance,
        #     min_node,
        #     max_node,
        # )

    def generate_run(self):
        """Generate the run configuration of the experiment.

        By definition, the run configuration is the combination of the training environment and the compute instance.

        Returns:
            RunConfiguration: The run configuration of the experiment.
        """

        # run_config = RunConfiguration()
        # docker_config = DockerConfiguration(use_docker=True)
        # run_config.docker = docker_config

        # aml_run_env = Environment.get(
        #     self.interface.workspace,
        #     self.env_name,
        # )

        # run_config.environment = aml_run_env

        # run_config.target = self.compute_target
        src_dir = str(Path.cwd())

        job = command(
            code=src_dir,
            command=self.training_script_path,
            environment=f"{self.env_name}@latest",
            compute=self.aml_compute_name,
            display_name=self.experiment_name,
        )

        return job

    # def submit_run(self, job):
    #     """Submit your training loop and create an experiment.

    #     This experiment is defined by the use of the `ScriptRunConfig` class.
    #     This means that you have to provide the path to a script defining your training loop, defined by the parameters
    #     `source_directory` and `script` of the `ScriptRunConfig` class, `script` being the path of your training loop
    #     relative to `source_directory`.

    #     For example purpose, a training loop example is provided [TrainingLoopExample][azure_helper.steps.train] is
    #     provided, as well as an abstract class if you want to use this training loop structure, but you're not forced to.
    #     """

    #     returned_job = self.aml_interface.create_or_update(job)
    #     aml_url = returned_job.studio_url
    #     print("Monitor your job at", aml_url)



from pathlib import Path
from typing import List

from azureml.core import Environment, Experiment, ScriptRunConfig
from azureml.core.runconfig import DockerConfiguration, RunConfiguration
from azureml.pipeline.core import Pipeline
from azureml.pipeline.steps import PythonScriptStep

from azure_helper.interfaces.aml_interface import AMLInterface
from azure_helper.logger import get_logger

log = get_logger()


class AMLExperiment:
    def __init__(
        self,
        aml_interface: AMLInterface,
        aml_compute_name: str,
        aml_compute_instance: str,
        env_name: str,
        experiment_name: str,
        training_script_path: str,
        min_node: int = 1,
        max_node: int = 2,
        clean_after_run: bool = True,
    ) -> None:
        """Instantiate the creation of an AzureML Experiment needed to train a model.

        Args:
            aml_interface (AMLInterface): The aml_interface needed to register the experiment in the workspace.
            aml_compute_name (str): The name of the compute instance used.
            aml_compute_instance (str): The size (as in `vm_size`) of the compute instance used.
            min_node (int, optional): The minimum number of nodes to use on the compute instance. Defaults to 1.
            max_node (int, optional): The maximum number of nodes to use on the compute instance. Defaults to 2.
            env_name (str): The name of the training environment.
            experiment_name (str): The name of the experiment.
            training_script_path (str): The path to the training loop script.
            clean_after_run (bool, optional): Whether or not you want to delete the compute instance after training. Defaults to True.
        """

        self.interface = aml_interface
        self.aml_compute_name = aml_compute_name
        self.aml_compute_instance = aml_compute_instance
        self.env_name = env_name
        self.experiment_name = experiment_name
        self.clean_after_run = clean_after_run
        self.training_script_path = training_script_path

        self.compute_target = aml_interface.get_compute_target(
            aml_compute_name,
            aml_compute_instance,
            min_node,
            max_node,
        )

    def generate_run_config(self) -> RunConfiguration:
        """Generate the run configuration of the experiment.

        By definition, the run configuration is the combination of the training environment and the compute instance.

        Returns:
            RunConfiguration: The run configuration of the experiment.
        """

        run_config = RunConfiguration()
        docker_config = DockerConfiguration(use_docker=True)
        run_config.docker = docker_config

        aml_run_env = Environment.get(
            self.interface.workspace,
            self.env_name,
        )

        run_config.environment = aml_run_env

        run_config.target = self.compute_target

        return run_config

    def submit_pipeline(self, steps: List[PythonScriptStep]):
        """Used to submit a training pipeline.

        Args:
            steps (List[PythonScriptStep]): The different steps of the training pipeline.

        https://github.com/Azure/MachineLearningNotebooks/blob/master/how-to-use-azureml/machine-learning-pipelines/intro-to-pipelines/aml-pipelines-getting-started.ipynb

        This should be a registered AZML Pipeline with the followings steps :

        * Download raw datas from one datastore
        * Transform those datas in "gold" datas and store them in another datastore
        * Use these gold datas to train a model
        * Evaluate that model
        * Save and version that model

        """
        experiment = Experiment(self.interface.workspace, self.experiment_name)
        # src_dir = __here__
        # src_dir = str(Path.cwd())

        pipeline = Pipeline(workspace=self.interface.workspace, steps=steps)

        # TODO: add the validation method of the pipeline.
        log.info("Submitting Run")
        run = experiment.submit(config=pipeline)
        run.wait_for_completion(show_output=True)
        log.info("Run completed.")

        if self.clean_after_run:
            log.info("Deleting compute instance.")
            self.compute_target.delete()

    def submit_run(self):
        """Submit your training loop and create an experiment.

        This experiment is defined by the use of the `ScriptRunConfig` class.
        This means that you have to provide the path to a script defining your training loop, defined by the parameters
        `source_directory` and `script` of the `ScriptRunConfig` class, `script` being the path of your training loop
        relative to `source_directory`.

        For example purpose, a training loop example is provided [TrainingLoopExample][azure_helper.steps.train] is
        provided, as well as an abstract class if you want to use this training loop structure, but you're not forced to.
        """

        experiment = Experiment(self.interface.workspace, self.experiment_name)
        # src_dir = __here__
        src_dir = str(Path.cwd())

        docker_config = DockerConfiguration(use_docker=True)
        script = ScriptRunConfig(
            source_directory=src_dir,
            script=self.training_script_path,
            docker_runtime_config=docker_config,
        )

        script.run_config = self.generate_run_config()

        # script.run_config.target = self.compute_target

        # aml_run_env = Environment.get(
        #     self.interface.workspace,
        #     self.env_name,
        # )
        # script.run_config.environment = aml_run_env

        log.info("Submitting Run")
        run = experiment.submit(config=script)
        run.wait_for_completion(show_output=True)
        log.info(f"Run completed : {run.get_metrics()}")

        if self.clean_after_run:
            log.info("Deleting compute instance.")
            self.compute_target.delete()



from pathlib import Path

from azureml.core.environment import Environment
from pydantic import BaseModel

from azure_helper.interfaces.aml_interface import AMLInterface
from azure_helper.logger import get_logger

log = get_logger()


class EnvSpecs(BaseModel):
    """Pydantic class used to specify how to build the environment : with pip, conda or docker.

    Args:
        flavor (str): Which method you chose, either pip, conda of docker.
        spec_file (Path): The path to either the `requirements.txt` files, the `environment.yml` file, or the `Dockerfile`.
    """

    flavor: str
    spec_file: Path


class AMLEnvironment:
    def __init__(self, private_wheel_location: Path, base_image: str) -> None:
        """Instantiate the creation of the AzureML Environment needed for the experiment.

        An AzureML Environment is a Docker Image which encapsulates all the needed requirements for the experiment (ie
        the training of a model) to run, this can be :

        * the version of the OS,
        * the various python libraries needed (ie the `requirements.txt`),
        * the project built as a wheel and added as a private pip package,
        * third-party softwares, etc.

        To create the Docker image of the environment, we use a base image managed by Microsoft. This image can be changed corresponding to the
        need of the experiment.

        The standard image we use is `mcr.microsoft.com/azureml/curated/sklearn-1.0-ubuntu20.04-py38-cpu`.

        The different images managed by Microsoft can be found ine the [Azure Machine Learning base images](https://github.com/Azure/AzureML-Containers).

        The create the Docker image of the environment, we also need the project, built as a pip wheel (by Flit, setuptools, Hatch, etc),
        so we can install it in our environment.

        Args:
            dist_dir (Path): The path where the project built as a wheel is located.
            base_image (str): The base Docker image we use to create the environment.
        """
        self.private_wheel_location = private_wheel_location
        self.base_image = base_image

    def validate_dir(self):
        """Small validation to check if a given path is a valid path in the project.

        Raises:
            FileNotFoundError: The path does not point to a valid directory.
        """
        if self.private_wheel_location.is_dir():
            log.info(f"Looking for wheel file in {self.private_wheel_location}.")
        else:
            raise FileNotFoundError

    def retrieve_whl_filepath(self) -> Path:
        """The project inside you develop your model you want to train need to be build as a pip wheel and added to the training environement.

        Once your project has been build as a pip wheel (by Flit, setuptools, Hatch, etc), given the location of the distribution directory
        (usually `$cwd/dist`), this function looks for a `.whl` file and return its path.

        !!! remark "Remark"

            Usually, there is only one `.whl` file in a `dist` directory. This is an assumption of this function.

        Raises:
            FileNotFoundError: Either the `dist` directory path is not valid, or there is no `.whl` file inside.

        Returns:
            Path: The path of the private wheel package.
        """
        try:
            self.validate_dir()
        except FileNotFoundError:
            log.error(
                f"Couldn't find distribution directory {self.private_wheel_location}",
            )
            raise

        whl_file = sorted(
            Path(file)
            for file in Path(self.private_wheel_location).glob("**/*.whl")
            if file.is_file()
        )
        if len(whl_file) == 0:
            log.error("Couldn't find wheel distribution")
            raise FileNotFoundError

        log.info(f"Found wheel {self.private_wheel_location / Path(whl_file[0])}")

        return self.private_wheel_location / Path(whl_file[0])

    def create_aml_environment(
        self,
        env_name: str,
        env_specs: EnvSpecs,
        aml_interface: AMLInterface,
    ) -> Environment:
        """Create the AzureML Environment once all the requirements have been gathered.

        Args:
            env_name (str): The name of the environment that will be build.
            env_specs (EnvSpecs): The specifications used to create the environement (ie pip, conda, or docker).
            aml_interface (AMLInterface): The AML interface which will be responsible to register the environment in the right workspace.

        Raises:
            ValueError: You have selected something else than "pip", "conda", or "docker" for the `EnvSpecs`.

        Returns:
            Environment: The training environment which will be used.

        !!! remark "Remark"
            The environment returned is just a Dockerfile, it not built as an image and will have to be built during the first training.
            If you want to build it locally, you will have to put `build_locally=True` in `AMLInterface.register_aml_environment`.
        """
        if env_specs.flavor == "pip":
            env = Environment.from_pip_requirements(
                name=env_name,
                file_path=str(env_specs.spec_file),
            )
        elif env_specs.flavor == "conda":
            env = Environment.from_conda_specification(
                name=env_name,
                file_path=str(env_specs.spec_file),
            )
        elif env_specs.flavor == "docker":
            env = Environment.from_dockerfile(
                name=env_name,
                dockerfile=str(env_specs.spec_file),
            )
        else:
            log.error(
                f"env_specs flavor {env_specs.flavor} is not a valid one. Only 'pip', 'conda', or 'docker' are valid choices.",
            )
            raise ValueError

        whl_filepath = self.retrieve_whl_filepath()
        private_wheel = env.add_private_pip_wheel(
            workspace=aml_interface.workspace,
            file_path=whl_filepath,
            exist_ok=True,
        )
        env.python.conda_dependencies.add_pip_package(private_wheel)

        env.docker.base_image = self.base_image

        # env.python.user_managed_dependencies = True
        # https://stackoverflow.com/questions/67387249/how-to-use-azureml-core-runconfig-dockerconfiguration-class-in-azureml-core-envi

        return env



from azure.ai.ml import MLClient
from azure.ai.ml.entities import AmlCompute, AzureBlobDatastore, Environment
from azure.identity import DefaultAzureCredential, ManagedIdentityCredential

from azure_helper.logger import get_logger

log = get_logger()


class AMLInterface:
    def __init__(
        self,
        managed_id: str,
        subscription_id: str,
        workspace_name: str,
        resource_group: str,
    ):
        """Instantiate the connection to an Azure Machine Learning Workspace.

        This class is the principal interface with the other ones. It uses a Service Principle **with Password Authentication** connection to interact with Azure
        Machine Learning. The Service Principle credentials needed can be encapsulated in the following Pydantic class.

        ```python
        from pydantic import BaseSettings, Field

        class SpnCredentials(BaseSettings):
            tenant_id: str = Field(env="TENANT_ID")
            service_principal_id: str = Field(env="SPN_ID")
            service_principal_password: str = Field(env="SPN_PASSWORD")
        ```

        It is responsible for :

        * Connect to an existing Storage Account and promote blob inside a specified container into an Azure Machine Learning Datastores.
        * Register Azure Machine Learning Environment created by the [`AMLEnvironment`][azure_helper.steps.create_aml_env] class as Docker Image.
        * Provisioning Compute Instance, either by fetching an existing one or creating a new one.


        Args:
            subscription_id (str): The Azure subscription ID containing the workspace.
            workspace_name (str): The workspace name. The name must be between 2 and 32 characters long.
                The first character of the name must be alphanumeric (letter or number), but the rest of the name may
                contain alphanumerics, hyphens, and underscores. Whitespace is not allowed.
            resource_group (str): The resource group containing the workspace.
        """
        auth = DefaultAzureCredential()
        # auth = ManagedIdentityCredential()
        self.workspace = MLClient(
            credential=auth,
            workspace_name=workspace_name,
            subscription_id=subscription_id,
            resource_group_name=resource_group,
        )

    def register_datastore(
        self,
        datastore_name: str,
        container_name: str,
        storage_acct_name: str,
        description: str,
    ):
        """Register an Azure Blob Container as an AZML datastore.

        ie if you have an architecture like the following one.

        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : train
            │   └── blob : test
        ```

        Then


        ```py
        aml_interface.register_datastore(
            datastore_name="project_name",
            container_name="project-mlops-mk-5448820782",
            storage_acct_name="workspaceperso5448820782",
            description="small description,
            )
        ```

        Will register the Blob container `project-mlops-mk-5448820782` as an AZML datastore under the name `project_name`.


        !!! attention "Attention"

            We are talking here about **datastores**, not **datasets**, which are special data assets of a datastore.

            Eg, in the following architecture.

            ```bash
            Storage_Account : workspaceperso5448820782
                │
                ├── Container : project-mlops-mk-5448820782
                │   ├── blob : train
                │   │           ├── x_train.csv
                │   │           └── y_train.csv
                │   └── blob : test
            ```

            `x_train.csv` and `y_train.csv` can be registered as datasets.

        Args:
            datastore_name (str): The name of the AZML datastore, case insensitive, can only contain alphanumeric characters and -.
            container_name (str): The name of the azure blob container.
            storage_acct_name (str): The storage account name.
            description (str): Small description of the datastore.
        """
        store = AzureBlobDatastore(
            name=datastore_name,
            container_name=container_name,
            account_name=storage_acct_name,
            description=description,
        )

        self.workspace.create_or_update(store)

    def register_aml_environment(
        self,
        environment: Environment,
    ):
        """Register the environment object in your workspace.

        An environement created by the [`AMLEnvironment`][azure_helper.steps.create_aml_env] class encapsulate in a Docker image
        everything which is needed to make to project work. See its documentation for further explainations.

        Args:
            environment (Environment): A reproducible Python environment for machine learning experiments.
        """

        self.workspace.environments.create_or_update(environment)

    def submit_experiment(self, job):
        """Register the environment object in your workspace.

        An environement created by the [`AMLEnvironment`][azure_helper.steps.create_aml_env] class encapsulate in a Docker image
        everything which is needed to make to project work. See its documentation for further explainations.

        Args:
            environment (Environment): A reproducible Python environment for machine learning experiments.
        """

        returned_job = self.workspace.create_or_update(job)
        aml_url = returned_job.studio_url
        log.info(f"Monitor your job at {aml_url}")

    def get_compute_target(
        self,
        compute_name: str,
        vm_size: str = "",
        min_node: int = 1,
        max_node: int = 2,
    ) -> AmlCompute:
        """Instantiate a compute instance to train the models.

        If no Compute Instance with the specified parameters is found in the workspace, a new one will be created.

        Args:
            compute_name (str): The name of the compute instance.
            vm_size (str): The size of agent VMs in the the compute instance.
                More details can be found [here](https://aka.ms/azureml-vm-details).
                Note that not all sizes are available in all regions, as detailed in the previous link.
                If not specified, (ie `vm_size = ""`) defaults to `Standard_NC6`.
            min_node (int, optional): The minimum number of nodes to use on the cluster. Defaults to 1.
            max_node (int, optional): The maximum number of nodes to use on the cluster. Defaults to 2.


        Returns:
            ComputeInstance: An instantiated compute instance.
        """
        compute_name = compute_name
        cluster_basic = AmlCompute(
            name=compute_name,
            size=vm_size,
            min_instances=min_node,
            max_instances=max_node,
        )

        self.workspace.begin_create_or_update(cluster_basic)

        log.info(
            f"Compute target instantiated : {self.workspace.compute.get(compute_name)}",
        )
        return cluster_basic



from io import StringIO

import pandas as pd
from azure.core.exceptions import ResourceExistsError
from azure.storage.blob import BlobServiceClient

from azure_helper.logger import get_logger

log = get_logger()


class BlobStorageInterface:
    def __init__(self, storage_acct_name: str, storage_acct_key: str):
        """Class responsible to interact with an existing Azure Storage Account.

        It uses a connection string to connect to the Storage Account.

        ```python
        conn_str = (
            "DefaultEndpointsProtocol=https;"
            + f"AccountName={storage_acct_name};"
            + f"AccountKey={storage_acct_key};"
            + "EndpointSuffix=core.windows.net"
        )
        ```

        !!! info "Information"

            To get the key of this storage account we use the following command with the azure-cli.

            ```sh
            az storage account keys list --account-name <storage-account-name> --resource-group <resource-group>
            ```

        This class is responsible for :

        * Creating a container in the storage account.
        * Uploading a dataframe (as a `csv` for now) inside a blob in one of the container of the storage account.
        * Download a `csv` from a blob in one of the container of the storage account and render it as a pandas dataframe.

        Args:
            storage_acct_name (str): The name of the storage account to which you want to connect.
            storage_acct_key (str): The account key of the storage account.
        """
        conn_str = (
            "DefaultEndpointsProtocol=https;"
            + f"AccountName={storage_acct_name};"
            + f"AccountKey={storage_acct_key};"
            + "EndpointSuffix=core.windows.net"
        )
        self.blob_service_client = BlobServiceClient.from_connection_string(
            conn_str,
        )

    def create_container(self, container_name: str):
        """Create a container inside the storage account.

        Args:
            container_name (str): the name of the container you want to create. This name can only contains
                alphanumeric numbers and dashes '-'.
        """
        try:
            self.blob_service_client.create_container(container_name)
            log.info(f"Creating blob storage container {container_name}.")
        except ResourceExistsError:
            log.warning(f"Blob storage container {container_name} already exists.")
            pass

    def upload_to_blob(
        self,
        dataset,
        container_name: str,
        blob_path: str,
    ):
        """Upload a dataset file inside a blob.

        Eg the following code.


        ```python
        from azure_helper.utils.blob_storage_interface import BlobStorageInterface

        blob_storage_interface = BlobStorageInterface(
            storage_acct_name="workspaceperso5448820782",
            storage_acct_key="XXXXX-XXXX-XXXXX-XXXX",
            )

        dataset = dataframe.to_csv(index=False, header=True).encode()

        blob_storage_interface.upload_to_blob(
            dataset=dataset,
            container_name="project-mlops-mk-5448820782",
            blob_path="raw/data.csv",
        )
        ```

        Upload the dataframe `dataframe`as `raw.csv`in the following way.


        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : raw
            │   │           └── data.csv
            │   └── blob : test
        ```

        !!! attention "Attention"

            As of now, there is no **data versioning**. Meaning that if the `blob_path` already exists, it will be
            overwritten with new datas.


        Args:
            dataset (Any): The datas you want to upload.
            container_name (str): The name of the container on which you want to upload the dataframe.
            blob_path (str): The path to the csv
        """
        self.create_container(container_name)

        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )

        try:
            blob_client.upload_blob(dataset)
            log.info(f"Dataset uploaded at blob path : {blob_path}.")
        except ResourceExistsError:
            log.warning(
                f"Blob path {blob_path} already contains datas. Now deleting old datas tu upload the new ones.",
            )
            blob_client.delete_blob()
            blob_client.upload_blob(dataset)
            log.info(f"New dataset uploaded at blob path : {blob_path}.")

    def upload_df_to_blob(
        self,
        dataframe: pd.DataFrame,
        container_name: str,
        blob_path: str,
    ):
        """Upload a pandas dataframe as a `csv` file inside a blob.

        Eg the following code.


        ```python
        from azure_helper.utils.blob_storage_interface import BlobStorageInterface

        blob_storage_interface = BlobStorageInterface(
            storage_acct_name="workspaceperso5448820782",
            storage_acct_key="XXXXX-XXXX-XXXXX-XXXX",
            )

        blob_storage_interface.upload_df_to_blob(
            dataframe=x_train,
            container_name="project-mlops-mk-5448820782",
            blob_path="train/x_train.csv",
        )
        blob_storage_interface.upload_df_to_blob(
            dataframe=x_train,
            container_name="project-mlops-mk-5448820782",
            blob_path="train/y_train.csv",
        )
        ```

        Upload the dataframes `x_train` and `y_train` as `x_train.csv` and `y_train.csv` in the following way.


        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : train
            │   │           ├── x_train.csv
            │   │           └── y_train.csv
            │   └── blob : test
        ```

        !!! attention "Attention"

            As of now, there is no **data versioning**. Meaning that if the `blob_path` already exists, it will be
            overwritten with new datas.


        Args:
            dataframe (pd.DataFrame): The dataframe you want to upload.
            container_name (str): The name of the container on which you want to upload the dataframe.
            blob_path (str): The path to the csv
        """
        log.warning(
            "The function 'upload_df_to_blob' will be deprecated in favour of a more generic version 'upload_to_blob' in the near future.",
        )
        self.create_container(container_name)

        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )

        try:
            blob_client.upload_blob(
                dataframe.to_csv(index=False, header=True).encode(),
            )
            log.info(f"Dataset uploaded at blob path : {blob_path}.")
        except ResourceExistsError:
            log.warning(
                f"Blob path {blob_path} already contains datas. Now deleting old datas tu upload the new ones.",
            )
            blob_client.delete_blob()
            blob_client.upload_blob(
                dataframe.to_csv(index=False, header=True).encode(),
            )
            log.info(f"New dataset uploaded at blob path : {blob_path}.")

    def download_from_blob(self, container_name: str, blob_path: str) -> StringIO:
        """Download a file a the given `blob_path` location and renders it as a StringIO buffer.

        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : train
            │   │           ├── x_train.csv
            │   │           └── y_train.csv
            │   └── blob : test
        ```

        ```python
        from azure_helper.utils.blob_storage_interface import BlobStorageInterface

        blob_storage_interface = BlobStorageInterface(
            storage_acct_name="workspaceperso5448820782",
            storage_acct_key="XXXXX-XXXX-XXXXX-XXXX",
            )

        df_buffer = blob_storage_interface.download_from_blob(
            container_name="project-mlops-mk-5448820782",
            blob_path="train/x_train.csv",
        )
        # now convert to csvt to read it.
        dataframe = pd.read_csv(df_buffer)
        ```

        Args:
            container_name (str): The name of the container.
            blob_path (str): The path to the `csv` file.

        Returns:
            StringIO: the file as a StringIO buffer.
        """

        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )
        stream = blob_client.download_blob()
        buffer = StringIO(stream.content_as_text())
        log.info(f"Download from {container_name} ended successfully.")
        return buffer

    def download_blob_to_df(self, container_name: str, blob_path: str) -> pd.DataFrame:
        """Download a `csv` file a the given `blob_path` location and renders it as a pandas datatrame.

        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : train
            │   │           ├── x_train.csv
            │   │           └── y_train.csv
            │   └── blob : test
        ```

        ```python
        from azure_helper.utils.blob_storage_interface import BlobStorageInterface

        blob_storage_interface = BlobStorageInterface(
            storage_acct_name="workspaceperso5448820782",
            storage_acct_key="XXXXX-XXXX-XXXXX-XXXX",
            )

        df = blob_storage_interface.download_blob_to_df(
            container_name="project-mlops-mk-5448820782",
            blob_path="train/x_train.csv",
        )
        ```

        Args:
            container_name (str): The name of the container.
            blob_path (str): The path to the `csv` file.

        Returns:
            pd.DataFrame: the `csv` file as a dataframe.
        """
        log.warning(
            "The function 'download_blob_to_df' will be deprecated in favour of a more generic version 'download_from_blob' in the near future.",
        )

        blob_client = self.blob_service_client.get_blob_client(
            container=container_name,
            blob=blob_path,
        )
        stream = blob_client.download_blob()
        buffer = StringIO(stream.content_as_text())
        dataframe = pd.read_csv(buffer)
        log.info(f"Download from {container_name} ended successfully.")
        return dataframe
